use std::{
    sync::mpsc::{self, Receiver, Sender, channel},
    thread::{self, JoinHandle},
};

mod camera;
mod color;
mod film;
mod geometry;
mod sampling;
mod scene;

pub use camera::Camera;
pub use film::Film;
pub use geometry::{Point, Shape, Transform, Vector};
pub use scene::Scene;

use color::Rgb;
use geometry::{Point2, Transform2, point2};
use sampling::PixelSample;

pub struct RenderSession {
    thread: RenderThread,
}

impl RenderSession {
    pub fn new(scene: Scene, camera: Camera, film: Film, samples: u32, seed: u32) -> Self {
        Self {
            thread: RenderThread::spawn(scene, camera, film, samples, seed),
        }
    }

    pub fn stop(self) -> thread::Result<RenderState> {
        self.thread.stop()
    }

    pub fn is_finished(&self) -> bool {
        self.thread.is_finished()
    }

    pub fn join(self) -> thread::Result<RenderState> {
        self.thread.join()
    }
}

pub enum RenderState {
    Stopped(Film),
    Finished(Film),
}

struct RenderThread {
    handle: JoinHandle<RenderState>,
    shutdown: Sender<()>,
}

impl RenderThread {
    fn spawn(scene: Scene, camera: Camera, film: Film, samples: u32, seed: u32) -> Self {
        let (tx, rx) = channel::<()>();
        let handle = thread::spawn(move || render_thread(scene, camera, film, samples, seed, rx));

        Self {
            handle,
            shutdown: tx,
        }
    }

    fn is_finished(&self) -> bool {
        self.handle.is_finished()
    }

    fn join(self) -> thread::Result<RenderState> {
        self.handle.join()
    }

    fn stop(self) -> thread::Result<RenderState> {
        let _ = self.shutdown.send(());
        self.handle.join()
    }
}

fn render_thread(
    scene: Scene,
    camera: Camera,
    mut film: Film,
    samples: u32,
    seed: u32,
    shutdown: Receiver<()>,
) -> RenderState {
    let mut sampler = film.sampler(samples, seed);
    let t = film.transform;
    let w = film.width as usize;

    for (i, mut c) in film.lanes_mut().into_iter().enumerate() {
        let mut rgb = Rgb::BLACK;

        let x = i % w;
        let y = i / w;

        for mut pixel_sample in sampler.pixel(x as u32, y as u32) {
            let p_film = get_camera_sample(x as u32, y as u32, &mut pixel_sample, t);
            let ray = camera.ray(p_film);

            if let Some(h) = scene.intersect(ray) {
                rgb += dir_color(h.normal);
            }

            use mpsc::TryRecvError as E;
            match shutdown.try_recv() {
                Ok(_) | Err(E::Disconnected) => return RenderState::Stopped(film),
                Err(E::Empty) => (),
            }
        }

        rgb /= samples as f32;
        for i in 0..3 {
            c[i] = rgb[i];
        }
    }

    RenderState::Finished(film)
}

fn get_camera_sample(x: u32, y: u32, sampler: &mut PixelSample, transform: Transform2) -> Point2 {
    transform * (point2(x as f32, y as f32) + sampler.get_2d().into())
}

fn dir_color(v: Vector) -> Rgb {
    (v.normalize() * 0.5 + Vector::splat(0.5)).into()
}
