use std::{mem, panic};

use mechthild_core::{Camera, Film, RenderSession, RenderState, Scene, Shape};
use numpy::{Ix3, PyArray};
use pyo3::prelude::*;

#[pyclass(name = "Camera")]
#[derive(Clone)]
struct PyCamera(Camera);

#[pymethods]
impl PyCamera {
    #[new]
    pub fn new(
        eye: (f32, f32, f32),
        center: (f32, f32, f32),
        up: (f32, f32, f32),
        fov: f32,
    ) -> Self {
        Self(Camera::new(eye.into(), center.into(), up.into(), fov))
    }

    pub fn __str__(&self) -> PyResult<String> {
        Ok(format!("{:#?}", self.0))
    }
}

#[pyclass(name = "Shape")]
#[derive(Clone)]
struct PyShape(Shape);

#[pymethods]
impl PyShape {
    #[staticmethod]
    pub fn sphere(center: (f32, f32, f32), radius: f32) -> Self {
        Self(Shape::sphere(center.into(), radius))
    }

    #[staticmethod]
    pub fn plane(normal: (f32, f32, f32), d: f32) -> Self {
        Self(Shape::plane(normal.into(), d))
    }

    pub fn __str__(&self) -> PyResult<String> {
        Ok(format!("{:?}", self.0))
    }
}

#[pyclass(name = "Session")]
struct PySession(Session);

enum Session {
    Running(RenderSession),
    Finished(Film),
    Stopped(Film),
    // i hate this
    Empty,
}

impl Default for Session {
    fn default() -> Self {
        Self::Empty
    }
}

impl Session {
    fn new(scene: Scene, camera: Camera, film: Film, samples: u32, seed: u32) -> Self {
        Self::Running(RenderSession::new(scene, camera, film, samples, seed))
    }

    fn take(&mut self) -> Option<RenderSession> {
        let session = mem::take(self);

        if let Self::Running(s) = session {
            Some(s)
        } else {
            *self = session;
            None
        }
    }
}

#[pymethods]
impl PySession {
    #[new]
    #[pyo3(signature = (objects, camera, width, height, samples, seed = 0))]
    pub fn new(
        objects: Vec<PyShape>,
        camera: &PyCamera,
        width: u32,
        height: u32,
        samples: u32,
        seed: u32,
    ) -> Self {
        let film = Film::new(width, height);
        let scene = Scene::new(objects.into_iter().map(|s| s.0).collect());

        Self(Session::new(scene, camera.0.clone(), film, samples, seed))
    }

    pub fn wait(&mut self) {
        if let Some(session) = self.0.take() {
            *self = match session.join() {
                Ok(RenderState::Finished(f)) => Self(Session::Finished(f)),
                Ok(RenderState::Stopped(f)) => Self(Session::Stopped(f)),
                Err(e) => panic::resume_unwind(e),
            }
        }
    }

    pub fn stop(&mut self) {
        if let Some(session) = self.0.take() {
            *self = match session.stop() {
                Ok(RenderState::Finished(f)) => Self(Session::Finished(f)),
                Ok(RenderState::Stopped(f)) => Self(Session::Stopped(f)),
                Err(e) => panic::resume_unwind(e),
            }
        }
    }

    pub fn get_output<'py>(&mut self, py: Python<'py>) -> Bound<'py, PyArray<f32, Ix3>> {
        self.wait();

        match &self.0 {
            Session::Finished(f) => PyArray::from_array(py, &f.data),
            Session::Stopped(f) => PyArray::from_array(py, &f.data),
            _ => unreachable!(),
        }
    }
}

#[pymodule]
fn mechthild(m: &Bound<'_, PyModule>) -> PyResult<()> {
    m.add_class::<PyCamera>()?;
    m.add_class::<PyShape>()?;
    m.add_class::<PySession>()?;
    Ok(())
}
